<?php

/**
 * JCH Optimize - Performs several front-end optimizations for fast downloads.
 *
 * @author    Samuel Marshall <samuel@jch-optimize.net>
 * @copyright Copyright (c) 2022 Samuel Marshall / JCH Optimize
 * @license   GNU/GPLv3, or later. See LICENSE file
 *
 *  If LICENSE file missing, see <http://www.gnu.org/licenses/>.
 */

namespace JchOptimize\Core;

use _JchOptimizeVendor\GuzzleHttp\Psr7\Uri;
use _JchOptimizeVendor\GuzzleHttp\Psr7\UriResolver;
use _JchOptimizeVendor\Joomla\DI\ContainerAwareInterface;
use _JchOptimizeVendor\Joomla\DI\ContainerAwareTrait;
use _JchOptimizeVendor\Joomla\Registry\Registry;
use _JchOptimizeVendor\Psr\Http\Message\UriInterface;
use JchOptimize\Core\FeatureHelpers\CdnDomains;

\defined('_JCH_EXEC') or exit('Restricted access');
class Cdn implements ContainerAwareInterface
{
    use ContainerAwareTrait;
    public string $scheme = '';
    protected ?array $domains = null;
    protected array $filePaths = [];
    protected ?array $cdnFileTypes = null;
    private Registry $params;
    private bool $enabled;

    public function __construct(Registry $params)
    {
        $this->params = $params;
        $this->enabled = (bool) $this->params->get('cookielessdomain_enable', '0');

        switch ($params->get('cdn_scheme', '0')) {
            case '1':
                $this->scheme = 'http';

                break;

            case '2':
                $this->scheme = 'https';

                break;

            case '0':
            default:
                $this->scheme = '';

                break;
        }
    }

    /**
     * Returns an array of file types that will be loaded by CDN.
     *
     * @return array $aCdnFileTypes Array of file type extensions
     */
    public function getCdnFileTypes(): array
    {
        if (\is_null($this->cdnFileTypes)) {
            $this->initialize();
        }

        return $this->cdnFileTypes;
    }

    /**
     * Returns array of default static files to load from CDN.
     *
     * @return array $aStaticFiles Array of file type extensions
     */
    public static function getStaticFiles(): array
    {
        return ['css', 'js', 'jpe?g', 'gif', 'png', 'ico', 'bmp', 'pdf', 'webp', 'svg'];
    }

    public function prepareDomain(string $domain): UriInterface
    {
        // If scheme not included then we need to add forward slashes to make UriInterfaces
        // implementations recognize the domain
        if (!\preg_match('#^(?:[^:/]++:|//)#', \trim($domain))) {
            $domain = '//'.$domain;
        }

        return (new Uri($domain))->withScheme($this->scheme);
    }

    /**
     * @return array|bool|mixed
     */
    public function loadCdnResource(UriInterface $uri, ?UriInterface $origPath = null): UriInterface
    {
        $domains = $this->getCdnDomains();
        if (empty($origPath)) {
            $origPath = $uri;
        }
        // if disabled or no domain is configured abort
        if (!$this->enabled || empty($domains)) {
            return $origPath;
        }
        // If file already loaded on CDN return
        if ($this->isFileOnCdn($uri)) {
            return $origPath;
        }
        // We're now ready to load path on CDN but let's remove query first
        $path = $uri->getPath();
        // If we haven't matched a cdn domain to this file yet then find one.
        if (!isset($this->filePaths[$path])) {
            $this->filePaths[$path] = $this->selectDomain($this->domains, $uri);
        }
        if ('' === (string) $this->filePaths[$path]) {
            return $origPath;
        }

        return $this->filePaths[$path];
    }

    public function getCdnDomains(): array
    {
        if (\is_null($this->domains)) {
            $this->initialize();
        }

        return $this->domains;
    }

    public function isFileOnCdn(UriInterface $uri): bool
    {
        foreach ($this->getCdnDomains() as $domainArray) {
            if ($uri->getHost() === $domainArray['domain']->getHost()) {
                return \true;
            }
        }

        return \false;
    }

    private function initialize()
    {
        $defaultFiles = self::getStaticFiles();
        $domainArray = [];
        $this->cdnFileTypes = [];
        if ($this->enabled) {
            if ('' != \trim($this->params->get('cookielessdomain', ''))) {
                $domain1 = $this->params->get('cookielessdomain');
                $sStaticFiles1 = \implode('|', \array_merge($this->params->get('staticfiles', $defaultFiles), $this->params->get('pro_customcdnextensions', [])));
                $domainArray['domain1']['domain'] = $this->prepareDomain($domain1);
                $domainArray['domain1']['extensions'] = $sStaticFiles1;
            }
            if (\JCH_PRO) {
                // @see CdnDomains::addCdnDomains()
                $this->container->get(CdnDomains::class)->addCdnDomains($domainArray);
            }
        }
        $this->domains = $domainArray;
        if (!empty($this->domains)) {
            foreach ($this->domains as $domains) {
                $this->cdnFileTypes = \array_merge($this->cdnFileTypes, \explode('|', $domains['extensions']));
            }
            $this->cdnFileTypes = \array_unique($this->cdnFileTypes);
        }
    }

    /**
     * @staticvar int $iIndex
     *
     * @return bool|string
     */
    private function selectDomain(array &$domainArray, UriInterface $uri): UriInterface
    {
        // If no domain is matched to a configured file type then we'll just return the file
        $cdnUri = new Uri();
        for ($i = 0; \count($domainArray) > $i; ++$i) {
            $domain = \current($domainArray);
            $staticFiles = $domain['extensions'];
            \next($domainArray);
            if (\false === \current($domainArray)) {
                \reset($domainArray);
            }
            if (\preg_match('#\\.(?>'.$staticFiles.')#i', $uri->getPath())) {
                // Prepend the cdn domain to the file path if a match is found.
                $cdnUri = UriResolver::resolve($domain['domain'], $uri->withScheme('')->withHost(''));

                break;
            }
        }

        return $cdnUri;
    }
}
