<?php

/**
 * JCH Optimize - Performs several front-end optimizations for fast downloads.
 *
 * @author    Samuel Marshall <samuel@jch-optimize.net>
 * @copyright Copyright (c) 2020 Samuel Marshall / JCH Optimize
 * @license   GNU/GPLv3, or later. See LICENSE file
 *
 * If LICENSE file missing, see <http://www.gnu.org/licenses/>.
 */

namespace JchOptimize\Platform;

\defined('_WP_EXEC') or exit('Restricted access');
use _JchOptimizeVendor\Joomla\Uri\Uri;
use JchOptimize\Core\Admin\AbstractHtml;
use JchOptimize\Core\Exception;
use JchOptimize\Core\FileUtils;

class Html extends AbstractHtml
{
    public function getHomePageHtml(): string
    {
        JCH_DEBUG ? \JchOptimize\Platform\Profiler::mark('beforeGetHtml') : null;
        $url = \home_url().'/?jchbackend=1';

        try {
            $response = $this->http->get($url, ['Accept-Encoding' => 'identity;q=0']);
            if (200 != $response->getStatusCode()) {
                throw new Exception\RuntimeException(\JchOptimize\Platform\Utility::translate('Failed fetching front end HTML with response code '.$response->getStatusCode()));
            }
            JCH_DEBUG ? \JchOptimize\Platform\Profiler::mark('afterGetHtml') : null;
            $body = $response->getBody();
            // Set pointer to beginning of stream.
            $body->rewind();

            return $body->getContents();
        } catch (\Exception $e) {
            $this->logger->error($url.': '.$e->getMessage());
            JCH_DEBUG ? \JchOptimize\Platform\Profiler::mark('afterGetHtml)') : null;

            throw new Exception\RuntimeException(\__('Load or refresh the front-end site first then refresh this page to populate the multi select exclude lists.'));
        }
    }

    public function getMainMenuItemsHtmls($iLimit = 5, $bIncludeUrls = \false): array
    {
        $aMenuUrls = $this->getMenuUrls();
        $aMenuUrls = \array_slice($aMenuUrls, 0, $iLimit);
        $aHtmls = [];
        // Limit the time spent on this
        $iTimer = \microtime(\true);
        foreach ($aMenuUrls as $menuUrl) {
            try {
                if ($bIncludeUrls) {
                    $aHtmls[] = ['url' => $menuUrl, 'html' => $this->getHtml($menuUrl)];
                } else {
                    $aHtmls[] = $this->getHtml($menuUrl);
                }
            } catch (Exception\ExceptionInterface $e) {
                $this->logger->error($e);
            }
            if (\microtime(\true) > $iTimer + 10.0) {
                break;
            }
        }

        return $aHtmls;
    }

    protected function getMenuUrls(): array
    {
        $homeUrl = \rtrim(\home_url(), '/\\');
        // Start array of oMenu urls with home page
        $aMenuUrls = [$homeUrl];
        $aMenus = \wp_get_nav_menus();
        // If nothing just work with the home url
        if (!$aMenus) {
            return $aMenuUrls;
        }
        $locations = \get_registered_nav_menus();
        $aMenuLocations = \get_nav_menu_locations();
        // Iterate through menus to find primary
        foreach ($aMenus as $oMenu) {
            if ($oMenu->term_id == $aMenuLocations['menu-primary']) {
                break;
            }
        }
        $aMenuItems = \wp_get_nav_menu_items($oMenu);
        foreach ($aMenuItems as $oMenuItem) {
            if (\rtrim($oMenuItem->url, '/\\') == $homeUrl) {
                continue;
            }
            if (!($cleanUrl = $this->cleanUrl($oMenuItem->url))) {
                continue;
            }

            /** @var FileUtils $fileUtils */
            $fileUtils = $this->getContainer()->get(FileUtils::class);
            if ($fileUtils->isInternal($cleanUrl)) {
                $aMenuUrls[] = $cleanUrl;
            }
        }

        return $aMenuUrls;
    }

    /**
     * @param mixed $sUrl
     *
     * @throws Exception\RuntimeException
     */
    protected function getHtml($sUrl): string
    {
        $oUri = new Uri($sUrl);
        $sQuery = $oUri->getQuery();
        \parse_str($sQuery, $aQuery);
        $aNewQuery = \array_merge($aQuery, ['jchbackend' => '1']);
        $oUri->setQuery($aNewQuery);
        $response = $this->http->get($oUri->toString(), ['Accept-Encoding' => 'identity;q=0']);
        if (200 != $response->getStatusCode()) {
            throw new Exception\RuntimeException('Failed fetching HTML: '.$sUrl.' - Response code: '.$response->getStatusCode());
        }
        $body = $response->getBody();
        $body->rewind();

        return $body->getContents();
    }

    private function cleanUrl(string $oMenuItem)
    {
        $oUri = new Uri($oMenuItem);

        return $oUri->toString(['scheme', 'host', 'path', 'query']);
    }
}
